package io.tswf.transmutations.view.generation

import groovy.transform.CompileStatic
import groovy.util.logging.Slf4j
import io.tswf.transmutations.common.conflict.ConflictResolver
import io.tswf.transmutations.view.exception.InvalidViewTargetException
import io.tswf.transmutations.view.exception.MethodMappingException
import io.tswf.transmutations.view.meta.ViewMetadata
import io.tswf.groovy.bettergroovy.transforms.validation.checkedmapconstructor.CheckedMapConstructor
import org.objectweb.asm.tree.ClassNode

@Slf4j
@CompileStatic
class ViewInjector {

    static void injectViews(ViewInjectionContext context) {
        def targetClass = context.targetClass
        def methodsForTarget = context.viewMeta.viewMethodsMetadata.filter { it.targetType.internalName == targetClass.name }

        if (methodsForTarget.isNullOrEmpty()) {
            throw new InvalidViewTargetException(context.viewMeta, targetClass)
        }

        for (def viewMethodMetadata : context.viewMeta.viewMethodsMetadata) {
            if (targetClass.interfaces == null) {
                targetClass.interfaces = new ArrayList<>()
            }

            // Проверяем, что такого метода еще нет в классе-цели
            def alreadyExistingConflictingMethod = targetClass.findSameMethod(viewMethodMetadata.viewMethod)
            if (alreadyExistingConflictingMethod != null) {
                def needsOverwriteMethod = ConflictResolver.resolveConflict(
                        targetClass: targetClass,
                        conflictingMethod: alreadyExistingConflictingMethod,
                        anotherMethod: viewMethodMetadata.viewMethod,
                        anotherMethodOwner: context.viewMeta.viewClass,
                        strategyProvider: viewMethodMetadata
                )

                if (needsOverwriteMethod) {
                    log.warn("Overwriting method ${alreadyExistingConflictingMethod.javaFullName} from ${targetClass.javaName} by view ${viewMethodMetadata.view.viewClass.javaName}...")
                    targetClass.methods.remove(alreadyExistingConflictingMethod)
                } else {
                    continue
                }
            }

            // Формируем контекст маппинга метода
            def viewMappingContext = new ViewMethodMappingContext(
                    viewMethodMeta: viewMethodMetadata,
                    targetClass: context.targetClass
            )

            try {
                viewMethodMetadata.mapper
                        .ensureNotNull { new RuntimeException("There is no mapper for view method ${viewMethodMetadata.viewMethod.javaFullName} in ${context.viewMeta.viewClass.javaName}") }
                        .handleViewMethodMapping(viewMappingContext)
            } catch (Throwable e) {
                throw new MethodMappingException(e, targetClass, viewMethodMetadata)
            }
        }

        // Добавляем интерфейс нашей фьюхи, если еще не добавлено. Дважды метод не добавит, так что вызываем без проверок.
        targetClass.addInterface(context.viewMeta.viewClass.asAsmType())
    }
}

@CheckedMapConstructor
@CompileStatic
class ViewInjectionContext {
    ViewMetadata viewMeta
    ClassNode targetClass
}
