package io.tswf.transmutations.view.generation.mapper

import groovy.transform.CompileStatic
import io.tswf.transmutations.view.exception.CantReachFieldException
import io.tswf.transmutations.view.exception.InvalidViewMethodException
import io.tswf.transmutations.view.generation.IViewMethodMapper
import io.tswf.transmutations.view.generation.ViewMethodMappingContext
import org.objectweb.asm.Type
import org.objectweb.asm.tree.FieldInsnNode
import org.objectweb.asm.tree.FieldNode
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.MethodNode
import org.objectweb.asm.tree.TypeInsnNode
import org.objectweb.asm.tree.VarInsnNode

import java.lang.reflect.Modifier

import static org.objectweb.asm.Opcodes.*

@CompileStatic
class FieldAccessorMapper implements IViewMethodMapper {

    String fieldName
    Type fieldType
    FieldAccessType accessType

    FieldAccessorMapper(String fieldName, Type fieldType, FieldAccessType accessType) {
        this.fieldName = fieldName
        this.fieldType = fieldType
        this.accessType = accessType
    }

    MethodNode generateGetter(FieldNode fieldToAccess, ViewMethodMappingContext context) {
        def viewMethod = context.viewMethodMeta.viewMethod
        def viewMethodParameterTypes = viewMethod.parameterTypes
        def viewMethodReturnType = viewMethod.returnType

        if (viewMethodParameterTypes.isNotEmpty()) {
            throw new InvalidViewMethodException("it must has no parameters as getter, but was ${viewMethodParameterTypes.map { it.className }}!", context.viewMethodMeta)
        }

        if (viewMethodReturnType != fieldType && !(!fieldType.isPrimitive() && viewMethodReturnType.descriptor != Object.asmDescriptor)) {
            throw new InvalidViewMethodException("it return type must match with target field type ${fieldType.className}, but was ${viewMethodReturnType.className}!", context.viewMethodMeta)
        }

        def newViewMethod = new MethodNode(ACC_PUBLIC, viewMethod.name, viewMethod.desc, viewMethod.signature, new String[0])

        def targetFieldIsStatic = Modifier.isStatic(fieldToAccess.access)

        // Если целевое поле - виртуальное, то нужно для доступа к нему загрузить this
        if (!targetFieldIsStatic) {
            newViewMethod.instructions << new VarInsnNode(ALOAD, 0)
        }

        newViewMethod.instructions << new FieldInsnNode(targetFieldIsStatic ? GETSTATIC : GETFIELD, context.targetClass.name, fieldName, fieldType.descriptor)
        newViewMethod.instructions << new InsnNode(fieldType.getOpcode(IRETURN))

        return newViewMethod
    }

    MethodNode generateSetter(FieldNode fieldToAccess, ViewMethodMappingContext context) {
        def viewMethod = context.viewMethodMeta.viewMethod
        def viewMethodParameterTypes = viewMethod.parameterTypes
        def viewMethodReturnType = viewMethod.returnType

        if (viewMethodReturnType != Type.VOID_TYPE) {
            throw new InvalidViewMethodException("it be void method, but was ${viewMethodReturnType.className}!", context.viewMethodMeta)
        }

        if (viewMethodParameterTypes.size() != 1) {
            throw new InvalidViewMethodException("it must has single parameter as setter, but was ${viewMethodParameterTypes.map { it.className }}!", context.viewMethodMeta)
        }

        def firstParameterType = viewMethodParameterTypes.first()

        // Записать мы можем или объект того же типа, или Object для не-примитивных полей на случай, если класс недоступен.
        if (firstParameterType != fieldType && !(!fieldType.isPrimitive() && firstParameterType.descriptor == Object.asmDescriptor )) {
            throw new InvalidViewMethodException("it first parameter type ${firstParameterType.className} doest not matches with target field type ${fieldType.className}", context.viewMethodMeta)
        }

        def newViewMethod = new MethodNode(ACC_PUBLIC, viewMethod.name, viewMethod.desc, viewMethod.signature, new String[0])

        def targetFieldIsStatic = Modifier.isStatic(fieldToAccess.access)

        // Если целевое поле - виртуальное, то нужно для доступа к нему загрузить this
        if (!targetFieldIsStatic) {
            newViewMethod.instructions << new VarInsnNode(ALOAD, 0)
        }

        // Задаем в поле мы первый параметр сеттера, у него всегда 1 индекс
        newViewMethod.instructions << new VarInsnNode(firstParameterType.getOpcode(ILOAD), 1)

        //TODO: Сделать красивый выбор исключения со словами: Ой, ошибка во время конвертации именно этой вьюхи

        // Для не примитивных типов делаем каст на всякий случай
        if (!fieldType.isPrimitive()) {
            newViewMethod.instructions << new TypeInsnNode(CHECKCAST, fieldType.internalName)
        }

        newViewMethod.instructions << new FieldInsnNode(targetFieldIsStatic ? PUTSTATIC : PUTFIELD, context.targetClass.name, fieldName, fieldType.descriptor)
        newViewMethod.instructions << new InsnNode(RETURN)

        return newViewMethod
    }

    @Override
    void handleViewMethodMapping(ViewMethodMappingContext context) {
        def targetClass = context.targetClass
        // Ищем поле, к которому требуется получить доступ
        def fieldToAccess = targetClass.fields.find { it.name == fieldName && it.desc == fieldType.descriptor }
            .ensureNotNull { new CantReachFieldException(fieldName, fieldType, targetClass) }

        // Если поле изначально final, то исправляем это недоразумение
        if (Modifier.isFinal(fieldToAccess.access)) {
            fieldToAccess.access &= ~Modifier.FINAL
        }

        def generatedAccessor = accessType == FieldAccessType.SET ? generateSetter(fieldToAccess, context) : generateGetter(fieldToAccess, context)
        targetClass.methods << generatedAccessor
    }
}

enum FieldAccessType {
    GET,
    SET
}
