package io.tswf.transmutations.view.generation.mapper

import groovy.transform.CompileStatic
import io.tswf.transmutations.common.remapping.TransmutationRemapper
import io.tswf.transmutations.view.exception.InvalidViewMethodException
import io.tswf.transmutations.view.generation.IViewMethodMapper
import io.tswf.transmutations.view.generation.ViewMethodMappingContext
import io.tswf.transmutations.view.util.ParametersHelper
import io.tswf.transmutations.view.util.ReturnTypeHelper
import org.objectweb.asm.Type
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.InsnNode
import org.objectweb.asm.tree.MethodInsnNode
import org.objectweb.asm.tree.MethodNode
import org.objectweb.asm.tree.VarInsnNode

import static org.objectweb.asm.Opcodes.ACC_PUBLIC
import static org.objectweb.asm.Opcodes.ALOAD
import static org.objectweb.asm.Opcodes.INVOKESPECIAL
import static org.objectweb.asm.Opcodes.IRETURN
import static org.objectweb.asm.Opcodes.POP
import static org.objectweb.asm.Opcodes.POP2

// Нарочно вынесено в отдельный класс на случай, если логика нормального вызова метода будет разрастаться
@CompileStatic
class SuperMethodInvokeMapper implements IViewMethodMapper {
    Method rawMethodToInvokeDescription

    SuperMethodInvokeMapper(Method rawMethodToInvokeDescription) {
        this.rawMethodToInvokeDescription = rawMethodToInvokeDescription
    }

    // Тут нет никаких проверок цели, так как это невозможно без загрузки новых ClassNode, что можно сделать не во всех средах. Unsafe.
    @Override
    void handleViewMethodMapping(ViewMethodMappingContext context) {
        if (rawMethodToInvokeDescription.name in MethodInvokeMapper.RESERVED_NAMES) {
            throw new InvalidViewMethodException("name '${rawMethodToInvokeDescription.name}' is reserved and can not be invoked via view", context.viewMethodMeta)
        }

        def targetSuperClass = Type.getObjectType(context.targetClass.superName)
        def mappedMethodToInvokeDescription = TransmutationRemapper.mapMethod(targetSuperClass, rawMethodToInvokeDescription)

        def viewMethod = context.viewMethodMeta.viewMethod
        def newMethod = new MethodNode(ACC_PUBLIC, viewMethod.name, viewMethod.desc, null, new String[0])
        def methodToInvokeReturnType = Type.getReturnType(mappedMethodToInvokeDescription.descriptor)

        // Сравниваем параметры для вызываемого метода и view метода. Делаем это через Helper для того, чтобы учитывать приведение типов с использованием сырого Object.
        ParametersHelper.validateMethodCallParameters(mappedMethodToInvokeDescription, context.viewMethodMeta)

        // Сравниваем возвращаемые типы. Делаем это через Helper для того, чтобы учитывать приведение типов с использованием сырого Object.
        def needsToPopMethodReturn = ReturnTypeHelper.validateMethodReturnTypesAndCheckPop(mappedMethodToInvokeDescription, context.viewMethodMeta)

        newMethod.instructions << new VarInsnNode(ALOAD, 0)

        // Загружаем параметры для вызова метода. Делаем это через Helper для того, чтобы учитывать приведение типов с использованием сырого Object.
        ParametersHelper.loadMethodCallParameters(newMethod.instructions, mappedMethodToInvokeDescription, context.viewMethodMeta)

        newMethod.instructions << new MethodInsnNode(INVOKESPECIAL, targetSuperClass.internalName, mappedMethodToInvokeDescription.name, mappedMethodToInvokeDescription.descriptor)

        // Если метод-цель отдает значение, а у view метода отдаваемый тип - void, то отданное значение надо удалить со стека
        if (needsToPopMethodReturn) {
            newMethod.instructions << new InsnNode(methodToInvokeReturnType.size == 1 ? POP : POP2)
        }

        newMethod.instructions << new InsnNode(viewMethod.returnType.getOpcode(IRETURN))

        context.targetClass.methods << newMethod
    }
}
