package io.tswf.transmutations.view.reading.mappers

import groovy.transform.CompileStatic
import io.tswf.transmutations.common.reader.MethodDescriptionReader
import io.tswf.transmutations.common.remapping.TransmutationRemapper
import io.tswf.transmutations.view.exception.InvalidViewMethodException
import io.tswf.transmutations.view.reading.IViewMethodMapperReader
import io.tswf.transmutations.view.reading.ViewMethodMapperReadingContext
import org.objectweb.asm.commons.Method
import org.objectweb.asm.tree.AnnotationNode

@CompileStatic
abstract class AbstractMethodInvokeMapperReader implements IViewMethodMapperReader {

    public static final String METHOD_DESCRIPTION_PARAMETER_VALUE = "value"

    Class annotationType
    Set<String> prefixes

    AbstractMethodInvokeMapperReader(Class annotationType, Set<String> prefixes) {
        this.prefixes = prefixes
        this.annotationType = annotationType
    }

    @Override
    void handleViewMethodReading(ViewMethodMapperReadingContext readingContext) {
        def viewMethod = readingContext.viewMethodMeta.viewMethod
        def methodInvokeAnnotation = viewMethod.findAnnotationByType(annotationType)
        def isViewMethodMatchesPrefixes = prefixes.any { prefix -> viewMethod.name.startsWith(prefix) && viewMethod.name.length() > prefix.length() }
        def methodIsInvocation = methodInvokeAnnotation != null || isViewMethodMatchesPrefixes

        if (methodIsInvocation) {
            def rawMethodName, rawMethodDesc

            def methodDescriptionAnnotation = methodInvokeAnnotation?.findValueByName(METHOD_DESCRIPTION_PARAMETER_VALUE, AnnotationNode)

            if (methodDescriptionAnnotation != null) {
                def methodDescription = MethodDescriptionReader.findMethodFromAnnotation(methodDescriptionAnnotation)
                rawMethodName = methodDescription.name
                rawMethodDesc = methodDescription.descriptor
            } else {
                rawMethodName = readingContext.viewMethodMeta.viewMethod.name.removeCamelCasePrefixes(prefixes, true)
                rawMethodDesc = viewMethod.desc
            }

            if (rawMethodName.isNullOrEmpty()) {
                throw new InvalidViewMethodException("extracted target method name is empty or null", readingContext.viewMethodMeta)
            }

            if (rawMethodDesc.isNullOrEmpty()) {
                throw new InvalidViewMethodException("extracted target method descriptor is empty or null", readingContext.viewMethodMeta)
            }

            // Проверки типов параметров очень громоздкие, так что реализованы только в самом маппере
            def rawMethod = new Method(rawMethodName, rawMethodDesc)

            // На момент чтения targetType еще имеет оригинальный тип до ремаппинга, так что валидно
            def mappedMethod = TransmutationRemapper.mapMethod(readingContext.viewMethodMeta.targetType, rawMethod)

            onMethodReaded(readingContext, mappedMethod)
        }
    }

    abstract void onMethodReaded(ViewMethodMapperReadingContext readingContext, Method methodDescription)
}
